/* $Id: lua_match.c,v 1.15 1996/03/29 16:29:10 miller Exp $ */

/* lua_match.c -- interface com lua da lib match */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "lua.h"
#include "pmatch.h"
#include "pm.h"

static lua_Object hook;

static void StoreSub (lua_Object table, char* index, long value)
{
  lua_pushobject(table);
  lua_pushliteral(index);
  lua_pushnumber(value);
  lua_storesubscript();
}

static int Capture (char chave, long start, long end, char *texto, void *user)
{
  lua_pushnumber (chave);
  lua_pushnumber (start+1);
  lua_pushnumber (end-start+1);
  lua_pushliteral (texto);
  lua_callfunction (hook);
  return 0;
}

static int Capture2 (char chave, long start, long end, char *texto, void *user)
{
  lua_Object newtable;
  long count;
  
  count = lua_getnumber(lua_getfield(hook, "count"));
  count++; 
  StoreSub(hook, "count", count);  

  newtable = lua_createtable();
  StoreSub(newtable, "index", chave);
  StoreSub(newtable, "start", start+1);
  StoreSub(newtable, "size", end - start + 1);  

  lua_pushobject(hook);
  lua_pushnumber(count);
  lua_pushobject(newtable);
  lua_storesubscript(); 
  
  return 0;
}

static void lua_match (void)
{
  lua_Object text, pattern, init;
  long length = 0, init1;
  char *text1, *pattern1;
  char usehook = 0;
  long first=0;

  if (((text = lua_getparam (1)) == LUA_NOOBJECT) ||
      ((pattern = lua_getparam (2)) == LUA_NOOBJECT))
  {
    lua_error ("match: too few arguments");
    return;
  }

  if ((init = lua_getparam (3)) == LUA_NOOBJECT)
    init1 = 0;
  else
    init1 = (long) lua_getnumber (init) - 1;

  if (!lua_isstring (text) || !lua_isstring (pattern))
    lua_error ("match: arguments of wrong type");

  text1 = lua_getstring (text);
  pattern1 = lua_getstring (pattern);

  if (text1[strlen(text1)] != '\0')
    strcat (text1, "\0");

  if ((hook = lua_getparam (4)) != LUA_NOOBJECT)
  {
    if (!lua_isfunction (hook))
      lua_error ("match: argument 4 should be a lua function");
  }
  else
  {
    usehook = 1;  
    hook = lua_createtable();
    StoreSub(hook, "count", 0);
    lua_pushobject(hook);
    lua_pushliteral("text");
    lua_pushliteral(text1);
    lua_storesubscript();
  }

  if (usehook)
    length = pm_pmatch (&first, text1, pattern1, init1, &Capture2, NULL);
  else
    length = pm_pmatch (&first, text1, pattern1, init1, &Capture, NULL);

  if (length > NOMATCH)
  {
    lua_pushnumber (first+1);
    lua_pushnumber (length);
    if (usehook)
      lua_pushobject(hook);
  }
}

static void lua_setchar (void)
{
  lua_Object temp1, temp2;
  int char_number;
  char *new_char;
  char result;		/* 0 == fail, 1 == OK */
  
  if (((temp1 = lua_getparam (1)) == LUA_NOOBJECT) ||
      ((temp2 = lua_getparam (2)) == LUA_NOOBJECT))
  {
    lua_error ("pm_setchar: too few arguments");
    return;
  };

  char_number = lua_getnumber (temp1);
  new_char    = lua_getstring (temp2);

  result = pm_setchar (char_number-1, *new_char);
  lua_pushnumber (result);
}

void pmlib_open ()
{
  lua_register ("match", lua_match);
	    /* Registra a funcao principal		    */
  lua_register ("pm_setchar", lua_setchar); 
	    /* Registra a funcao de mudanca de caracteres   */
  
  lua_pushnumber (0);
  lua_storeglobal ("pm_loop");		/* Valor default - nao faz loop	*/
}
