/* $Id: parse.c,v 1.9 1996/01/25 12:29:52 almendra Exp $ */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <setjmp.h>

#include "match.h"
#include "pmatch.h"
#include "parse.h"

    /* tabela de comandos e operadores */

extern function_st functions[MAX_FUNCTIONS];

int pm_setchar (int char_number, char new_char)
{
  if ((char_number < 0) || (char_number > strlen (pm_chars)))
    return 0;
  pm_chars[char_number] = new_char;

  return 1;
}

    /*
     * Funcao de interface, recebe padrao e retorna array contendo as finite
     * state machines (fsm) construidas a partir do padrao
     */

int pm_parse (char *text, fsm_st ** fsm)
{
  int t;
  vars_st vars;

  /* inicializacao das variaveis */

  vars.state = 1;
  vars.j = 0;
  vars.num_states = 0;
  vars.size = 0;
  vars.nextcap = 0;
  vars.string = text;

  if ((vars.fsm = (fsm_st *) malloc (STATE_BLOCK * sizeof (fsm_st))) == NULL)
    return MEM_ERROR;

  vars.num_states = STATE_BLOCK;

  /* a principio, nao ha captura. Se ocorrer uma, ele e setado
     para CAPTURE */

  vars.fsm[0].ch = NOCAPTURE;

  if (setjmp (vars.env) == 0)
    t = expression (&vars);

  else
  {
    free (vars.fsm);
    return PARSE_ERROR;
  }

  /* seta os estados inicial e final, guardando no inicial
     (fsm[0].next1) o tamanho da maquina */

  setstate (&vars, 0, vars.fsm[0].ch, NULL_CMD, t, vars.state + 1);
  setstate (&vars, vars.state, 0, NULL_CMD, 0, 0);

  *fsm = vars.fsm;

  return PARSE_OK;
}

static int expression (vars_st * vars)
{
  int r, t1;
  int last_state = vars->state - 1;

  t1 = term (vars);
  r = t1;

  if (vars->string[vars->j] == OR_CH)
  {
    int t2 = vars->state;
    int t3;

    r = t2;

    vars->j++;
    newstate (vars);

    t3 = expression (vars);	/* pega o 2o ramo do OR */

    /* faz o primeiro state antes do OR apontar para o state de entrada
     * do OR */

    if (vars->fsm[last_state].next1 == t1)
      vars->fsm[last_state].next1 = t2;

    if (vars->fsm[last_state].next2 == t1)
      vars->fsm[last_state].next2 = t2;

    /* faz o ultimo state do primeiro ramo do OR apontar para o 
     * state de saida do OR */

    if (vars->fsm[t2 - 1].next1 == t2)
      vars->fsm[t2 - 1].next1 = vars->state;

    if (vars->fsm[t2 - 1].next2 == t2)
      vars->fsm[t2 - 1].next2 = vars->state;

    setstate (vars, t2, 0, NULL_CMD, t1, t3);
    setstate (vars, vars->state, 0, NULL_CMD, vars->state + 1,
	      vars->state + 1);

    newstate (vars);
  }
  return r;
}

static int term (vars_st * vars)
{
  int r;

  r = factor (vars);

  if ((vars->string[vars->j] == OPGR_CH) ||
      (isvalid (vars->string[vars->j])) ||
      (vars->string[vars->j] == OPCL_CH) ||
      (vars->string[vars->j] == CAP_OPEN_CH) ||
      (vars->string[vars->j] == NEG_CH))
    term (vars);

  if (!((vars->string[vars->j] == OR_CH) ||
	(vars->string[vars->j] == CLGR_CH) ||
	(vars->string[vars->j] == '\0') ||
	(vars->string[vars->j] == CAP_CLOSE_CH)))
    error (vars);

  return r;
}

static int factor (vars_st * vars)
{
  int r, t1, t2 = 0;

  t1 = vars->state;

  if (vars->string[vars->j] == OPGR_CH)
  {
    vars->j++;
    t2 = expression (vars);

    if (vars->string[vars->j] == CLGR_CH)
      vars->j++;
    else
      error (vars);
  }

  else if (vars->string[vars->j] == CAP_OPEN_CH)
  {
    vars->fsm[0].ch = CAPTURE;
    setstate (vars, vars->state, vars->nextcap, CAP_OPEN_CMD,
	      vars->state + 1, vars->state + 1);
    t2 = vars->state;
    newstate (vars);
    vars->capture[++vars->size] = vars->nextcap++;
    vars->j++;

    expression (vars);

    if (vars->string[vars->j] == CAP_CLOSE_CH)
    {
      setstate (vars, vars->state, vars->capture[vars->size--],
		CAP_CLOSE_CMD, vars->state + 1, vars->state + 1);

      newstate (vars);
      vars->j++;
    }
    else
      error (vars);

  }

  else if (vars->string[vars->j] == ANY_CH)
  {
    setstate (vars, vars->state, 1, ANY_CMD, vars->state + 1,
	      vars->state + 1);
    t2 = vars->state;
    vars->j++;
    newstate (vars);
  }

  else if (vars->string[vars->j] == NEG_CH)
  {
    int t666;

    t2 = vars->state;
    vars->j++;
    newstate (vars);
    t666 = factor (vars);
    setstate (vars, t2, 1, NEG_OPEN_CMD, t666,
	      vars->state);
    setstate (vars, vars->state, 1, NEG_CLOSE_CMD, vars->state + 1,
	      vars->state + 1);
    newstate (vars);
  }

  else if (vars->string[vars->j] == BEGIN_CH)
  {
    setstate (vars, vars->state, 1, BEGIN_CMD, vars->state + 1,
	      vars->state + 1);
    t2 = vars->state;
    vars->j++;
    newstate (vars);
  }
  else if (vars->string[vars->j] == END_CH)
  {
    setstate (vars, vars->state, 1, END_CMD, vars->state + 1,
	      vars->state + 1);
    t2 = vars->state;
    vars->j++;
    newstate (vars);
  }

  else if (isvalid (vars->string[vars->j]) && (vars->string[vars->j]
			    != SPC_CH) && (vars->string[vars->j] != ANY_CH))
  {
    setstate (vars, vars->state, vars->string[vars->j], CHAR_CMD,
	      vars->state + 1, vars->state + 1);
    t2 = vars->state;
    vars->j++;
    newstate (vars);
  }

  else if (vars->string[vars->j] == OPCL_CH)
  {
    vars->j++;
    setstate (vars, vars->state, 0, CLASS_CMD, 0, 0);

    if (vars->string[vars->j] == NEGCL_CH)
    {
      vars->fsm[vars->state].next2 = 1;
      vars->j++;
    }

    t2 = vars->state;
    newstate (vars);

    if (vars->string[vars->j] == SEPCL_CH)
      error (vars);

    while ((vars->string[vars->j] != CLCL_CH) && (vars->string[vars->j] != '\n')
	   && (vars->string[vars->j] != '\0'))
    {
      if (vars->string[vars->j] == SPC_CH)
      {
	char temp;

	vars->j++;
	switch (vars->string[vars->j])
	{
	case 'n':
	  temp = '\n';
	  break;

	case 't':
	  temp = '\t';
	  break;

	case 'e':
	  temp = 27;
	  break;

	default:
	  temp = vars->string[vars->j];
	}
	setstate (vars, vars->state, temp, CLASS_CMD_CHAR, 0, 0);
	vars->j++;
	newstate (vars);
      }
      else if (vars->string[vars->j] == SEPCL_CH)
      {
	char temp = 0;

	vars->j++;

	if (vars->string[vars->j] == SPC_CH)
	{
	  vars->j++;
	  switch (vars->string[vars->j])
	  {
	  case 'n':
	    temp = '\n';
	    break;

	  case 't':
	    temp = '\t';
	    break;

	  case 'e':
	    temp = 27;
	    break;

	  default:
	    temp = vars->string[vars->j];
	  }
	}
	else if (vars->string[vars->j] != CLCL_CH)
	  temp = vars->string[vars->j];

	else
	  error (vars);

	setstate (vars, vars->state - 1,
		  vars->fsm[vars->state - 1].ch,
		  CLASS_CMD_RANGE, temp, 0);
	vars->j++;
      }
      else if (vars->string[vars->j] == BEGIN_CH)
      {
	setstate (vars, vars->state, 1, BEGIN_CMD, vars->state + 1,
		  vars->state + 1);
	t2 = vars->state;
	vars->j++;
	newstate (vars);
      }
      else if (vars->string[vars->j] == END_CH)
      {
	setstate (vars, vars->state, 1, END_CMD, vars->state + 1,
		  vars->state + 1);
	t2 = vars->state;
	vars->j++;
	newstate (vars);
      }

      else
      {
	setstate (vars, vars->state, vars->string[vars->j],
		  CLASS_CMD_CHAR, 0, 0);
	vars->j++;
	newstate (vars);
      }

    }
    if (vars->string[vars->j] != CLCL_CH)
      error (vars);
    setstate (vars, vars->state, 0, NULL_CMD, vars->state + 1,
	      vars->state + 1);
    vars->fsm[t2].next1 = vars->state;
    vars->j++;
    newstate (vars);

  }

  else if (vars->string[vars->j] == SPC_CH)
  {
    int loop1 = 0;

    vars->j++;

    while (functions[loop1].ch != '\0' && functions[loop1].ch
	   != vars->string[vars->j])
      loop1++;

    if (functions[loop1].ch == '\0')
    {
      int temp;

      switch (vars->string[vars->j])
      {
      case 'n':
	temp = '\n';
	break;

      case 't':
	temp = '\t';
	break;

      case 'e':
	temp = 27;
	break;

      case 'x':
	vars->j++;
	sscanf (&vars->string[vars->j], "%2x", &temp);
	vars->j++;
	break;

      case 'o':
	vars->j++;
	sscanf (&vars->string[vars->j], "%3o", &temp);
	vars->j += 2;
	break;

      default:
	if (isdigit (vars->string[vars->j]))
	{
	  sscanf (&vars->string[vars->j], "%3d", &temp);
	  if (temp > 255)
	  {
	    printf ("EE");
	    error (vars);
	  }
	  vars->j += 2;
	}
	else
	  temp = vars->string[vars->j];
      }
      setstate (vars, vars->state, (char) temp, CHAR_CMD,
		vars->state + 1, vars->state + 1);
    }

    else
    {
      setstate (vars, vars->state, loop1, SPC_CMD, vars->state + 1,
		vars->state + 1);
    }

    t2 = vars->state;
    vars->j++;
    newstate (vars);
  }
  else
    error (vars);

  if (vars->string[vars->j] == CL_CH)
  {
    setstate (vars, vars->state, 0, NULL_CMD, vars->state + 1, t2);
    r = vars->state;

    if (vars->fsm[t1 - 1].next1 == t1)
      vars->fsm[t1 - 1].next1 = vars->state;

    if (vars->fsm[t1 - 1].next2 == t1)
      vars->fsm[t1 - 1].next2 = vars->state;

    vars->j++;
    newstate (vars);
  }

  else if (vars->string[vars->j] == ONE_CH)
  {
    setstate (vars, vars->state, 0, NULL_CMD, vars->state + 1, t2);
    r = vars->state;

    if (vars->fsm[t1 - 1].next1 == t1)
      vars->fsm[t1 - 1].next1 = vars->state;

    if (vars->fsm[t1 - 1].next2 == t1)
      vars->fsm[t1 - 1].next2 = vars->state;

    if (vars->fsm[vars->state - 1].next1 == vars->state)
      vars->fsm[vars->state - 1].next1 = vars->state + 1;

    if (vars->fsm[vars->state - 1].next2 == vars->state)
      vars->fsm[vars->state - 1].next2 = vars->state + 1;

    vars->j++;
    newstate (vars);

    setstate (vars, vars->state, 0, NULL_CMD, vars->state + 1,
	      vars->state + 1);

    newstate (vars);
  }
  else if (vars->string[vars->j] == OOM_CH)
  {
    setstate (vars, vars->state, 0, NULL_CMD, vars->state + 1, t2);
    r = t2;

    vars->j++;

    newstate (vars);

  }
  else
    r = t2;

  return r;
}

static void error (vars_st * vars)
{
  longjmp (vars->env, 1);
}

/*************************************************/
/* Cria novo state e aloca memoria se necessario */
/*************************************************/

static void newstate (vars_st * vars)
{

  if (vars->state >= vars->num_states - 1)
  {
    fsm_st *new_fsm;

    new_fsm = (fsm_st *)
      realloc (vars->fsm, (vars->num_states + STATE_BLOCK)
	       * sizeof (fsm_st));

    if (new_fsm == NULL)
      exit (1);

    vars->fsm = new_fsm;
    vars->num_states += STATE_BLOCK;
  }

  vars->state++;
}

static void setstate (vars_st * vars, int state, char ch, char command,
		      int next1, int next2)
{
  vars->fsm[state].ch = ch;
  vars->fsm[state].command = command;
  vars->fsm[state].next1 = next1;
  vars->fsm[state].next2 = next2;
}
