/* $Id: rmatch.c,v 1.26 1996/01/30 12:43:19 miller Exp miller $ */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "pmatch.h"
#include "match.h"
#include "rmatch.h"

#define MAX_LAST_PATTERN 200
#define MIN_STACK_ELEMENTS 100
#define MIN_LIST_SIZE 10

  /* macro para simplificar funcoes do pattern matching */

#define isalphanum(x) (isalnum(x) || (x == '_'))

  /* variavel que armazena capturas temporariamente */

static capture_list_st capture_list =
{NULL, 0, 0, 0};

  /* indica se a rotina de pattern matching foi chamada pela
     funcao do usuario */

static enum bool
{
  false = 0, true = 1
}
will_be_recursive = false;

/*****************************************************************
 * Funcoes correspondentes aos comandos das expressoes regulares *
 *****************************************************************/

static match_t match_blanks (char *text, long j)
{
  return (((text[j] == '\t') || (text[j] == '\xff') || (text[j] == ' ') ||
	   (text[j] == '\n'))) ? normal_match : no_match;
}

static match_t match_non_blanks (char *text, long j)
{
  return (!((text[j] == '\t') || (text[j] == '\xff') ||
	    (text[j] == ' ') || (text[j] == '\n')))
    ? normal_match : no_match;
}

static match_t match_alpha (char *text, long j)
{
  return (isalpha (text[j])) ? normal_match : no_match;
}

static match_t match_non_alpha (char *text, long j)
{
  return (!isalpha (text[j]) && (text[j] != '\0'))
    ? normal_match : no_match;
}

static match_t match_digit (char *text, long j)
{
  return (isdigit (text[j])) ? normal_match : no_match;
}

static match_t match_non_digit (char *text, long j)
{
  return (!isdigit (text[j]) && (text[j] != '\0'))
    ? normal_match : no_match;
}

static match_t match_alphanum (char *text, long j)
{
  return isalphanum (text[j]) ? normal_match : no_match;
}

static match_t match_non_alphanum (char *text, long j)
{
  return (isalphanum (text[j]) || (text[j] == '\0'))
    ? no_match : normal_match;
}

static match_t match_word_boundary (char *text, long j)
{
  if ((j == 0) && isalphanum (text[j]))
    return no_char_match;

  else if (isalphanum (text[j - 1]) && !isalphanum (text[j]))
    return no_char_match;

  else if (isalphanum (text[j]) && !isalphanum (text[j - 1]))
    return no_char_match;

  return no_match;

}

/* array com os comandos do pattern matching com os 
 * respectivos ponteiros de funcao 
 */

function_st functions[] =
{
  {'w', &match_alphanum},
  {'W', &match_non_alphanum},
  {'d', &match_digit},
  {'D', &match_non_digit},
  {'S', &match_non_blanks},
  {'s', &match_blanks},
  {'b', &match_word_boundary},
  {'l', &match_alpha},
  {'L', &match_non_alpha},
  {'\0', NULL}
};

/**************************************************
 * Chama a funcao do usuario para cada captura ou *
 * organiza um array com elas caso o usuario nao  *
 * passado nenhuma                                * 
 **************************************************/

static void Capture_result (vars_st * vars, capt * capture)
{
  capt *next = NULL;

  capture_list.size = 0;

  while (capture != NULL)
  {
    capt *cap = capture->next_one;

    capture->next_one = next;
    next = capture;
    capture = cap;
  }

  capture = next;
  next = NULL;

  while (capture != NULL)
    if (capture->type == open)
    {
      capt *cap = capture->next_one;

      capture->next_one = next;
      next = capture;
      capture = cap;
    }

    else
    {
      if (capture->pos >= next->pos)
	(*vars->function) (capture->which_one, next->pos, capture->pos,
			   vars->text, vars->user);

      next = next->next_one;
      capture = capture->next_one;
    }
}

/*********************************************
 * funcao recursiva para percorrer a maquina *
 *********************************************/

static long rmatch (vars_st * vars, long j, int state,
		    capt * capture, int size)
{

  switch (vars->fsm[state].command)
  {
  case NULL_CMD:
    if (vars->fsm[state].next1 == 0)	/*se chegou ao fim da maquina de estados */
    {
      if (vars->function != NULL)
	Capture_result (vars, capture);		/* guarda capturas */

      return j;
    }

    /* verifica o estado atual ja foi avaliado antes */

    {
      int count;

      for (count = 0; count < size; count++)
	if (vars->tested[count] == state)
	  return NOMATCH;
    }

    vars->tested[size++] = state;	/* indicada que o estado foi testado */

    /* se houverem dois ramos, chama a funcao recursivamente,
       retornando com o primeiro que completar a maquina */

    if (vars->fsm[state].next1 != vars->fsm[state].next2)
    {
      long a;

      a = rmatch (vars, j, vars->fsm[state].next2, capture, size);

      if (a != NOMATCH)		/* se deu match */
	return a;

      a = rmatch (vars, j, vars->fsm[state].next1, capture, size);

      return a;
    }
    break;

  case CAP_OPEN_CMD:
    {
      long a;
      capt new_cap;

      new_cap.next_one = capture;
      new_cap.type = open;
      new_cap.pos = j;
      new_cap.which_one = vars->fsm[state].ch;

      a = rmatch (vars, j, vars->fsm[state].next1, &new_cap, size);

      return a;
    }
    break;

  case CAP_CLOSE_CMD:
    {
      long a;
      capt new_cap;

      new_cap.next_one = capture;
      new_cap.type = close;
      new_cap.pos = j - 1;
      new_cap.which_one = vars->fsm[state].ch;

      a = rmatch (vars, j, vars->fsm[state].next1, &new_cap, size);

      return a;
    }

  case CLASS_CMD:
    {
      int temp, found = 0, negate;

      temp = vars->fsm[state].next1;
      negate = vars->fsm[state].next2;
      state++;

      while (vars->fsm[state].command != NULL_CMD)
      {
	if (vars->fsm[state].command == CLASS_CMD_RANGE)
	{
	  if ((vars->text[j] >= vars->fsm[state].ch) &&
	      (vars->text[j] <= vars->fsm[state].next1))
	  {
	    found = 1;
	    break;
	  }
	}

	else if ((vars->fsm[state].command == CLASS_CMD_CHAR) &&
		 (vars->text[j] == vars->fsm[state].ch))
	{
	  found = 1;
	  break;
	};
	state++;
      }

      if (found ^ negate)
      {
	if (vars->text[j] == '\0')
	  return NOMATCH;
	j++;
	vars->tested = &vars->tested[size + 1];
	size = 0;
	return rmatch (vars, j, temp, capture, size);
      }
      else
	return NOMATCH;
    }

  case CHAR_CMD:
    if (vars->text[j] != vars->fsm[state].ch)
      return NOMATCH;
    j++;
    vars->tested = &vars->tested[size + 1];
    size = 0;
    break;

  case ANY_CMD:
    if ((vars->text[j] == '\0') || (vars->text[j] == '\n'))
      return NOMATCH;
    j++;
    vars->tested = &vars->tested[size + 1];
    size = 0;
    break;

  case SPC_CMD:
    {
      long a;

      a = (*functions[(int) vars->fsm[state].ch].function) (vars->text, j);

      switch (a)
      {
      case no_match:
	return NOMATCH;

      case normal_match:
	j++;
	vars->tested = &vars->tested[size + 1];
	size = 0;
	break;

      case no_char_match:;
	/* nao faz nada, continua normalmente */
      }
    }
    break;

  case BEGIN_CMD:
    if (!((vars->text[j - 1] == '\n') || (j == 0)))
      return NOMATCH;

    break;

  case END_CMD:
    if (!((vars->text[j] == '\n') || (vars->text[j] == '\0')))
      return NOMATCH;

    break;

/*  case NEG_OPEN_CMD:
   {
   long a;

   a = rmatch (vars, j, vars->fsm[state].next1, capture, size);

   if (a == NOMATCH)
   state = vars->fsm[state].next2;

   else
   return NOMATCH;

   break;
   }

   case NEG_CLOSE_CMD:
   return j + 1; */
  }

  return rmatch (vars, j, vars->fsm[state].next1, capture, size);
}

/*******************************************************
 * Guarda as capturas na estrutura capture_list, para  *
 * que o usuario possa acessa-las sem ter que fornecer * 
 * uma funcao para tal                                 *
 *******************************************************/

int pm_storeCaptures (char which_one, long start, long end,
		      char *text, void *user)
{
  if (capture_list.max_size == 0)
  {
    if ((capture_list.captures = (struct Capture *)
	 malloc (MIN_LIST_SIZE * sizeof (struct Capture))) == NULL)
        exit (1);

    capture_list.max_size = MIN_LIST_SIZE;
  }
  else if (capture_list.size == capture_list.max_size)
  {
    struct Capture *new_capture;

    if ((new_capture = (struct Capture *)
	 realloc (capture_list.captures,
		  capture_list.max_size * 2 * sizeof (struct Capture)))
	== NULL)
        exit (1);

    capture_list.captures = new_capture;
    capture_list.max_size *= 2;
  }

  capture_list.captures[capture_list.size].which_one = which_one;
  capture_list.captures[capture_list.size].start = start;
  capture_list.captures[capture_list.size].size = end-start+1;
  capture_list.size++;

  return 0;
}

/************************************************
 * verifica se um estado esta presente na pilha *
 ************************************************/

static int is_present (Stack * stack, int state)
{
  int a;

  for (a = 0; a < stack->size; a++)
    if (stack->stack[a] == state)
      return 1;

  return 0;
}

/************************************
 * Prepara uma pilha para ser usada *
 ************************************/

static int newStack (Stack * new_stack)
{
/*
   new_stack->stack = malloc (MIN_STACK_ELEMENTS * sizeof (new_stack->stack));

   if (new_stack->stack == NULL)
   return MEM_ERROR;
 */
  new_stack->size = 0;
  new_stack->max_size = MIN_STACK_ELEMENTS;

  return NORMAL;
}

/************************************
 * Coloca um novo elemento na pilha *
 ************************************/

static int push (Stack * stack, short value)
{
/*
   if (stack->size == stack->max_size)
   {
   short *new_array;

   new_array = (short *)
   realloc (stack->stack, sizeof (short) * stack->max_size * 2);

   if (new_array == NULL)
   return MEM_ERROR;

   stack->stack = new_array;
   stack->max_size *= 2;
   }
 */
  stack->stack[stack->size++] = value;

  return NORMAL;
}

/************************************** 
 * Libera memoria armazenada na pilha *
 **************************************/

static void clearStack (Stack * stack)
{
/*
   free (stack->stack);
 */
}

/*********************************************************** 
 * Copia a pilha fonte na pilha destino, apagando a fonte *
 ***********************************************************/

static void moveStack (Stack * dest, Stack * source)
{
   short *temp = dest->stack;
   short max_size = dest->max_size;

   dest->stack = source->stack;
   source->stack = temp;

   dest->size = source->size;
   source->size = 0;

   dest->max_size = source->max_size;
   source->max_size = max_size;
/*
  memcpy (dest, source, source->size * sizeof (short));

  dest->size = source->size;
  source->size = 0;
*/
}

/************************
 * Funcao nao recursiva * 
 ************************/

static long nmatch (char *text, fsm_st * fsm, long start)
{
  int finished = NOMATCH;
  Stack now, next;
  short a1[MIN_DEQUE_ELEMENTS];
  short a2[MIN_DEQUE_ELEMENTS];
  int state;
  int j = 0;
  int pos;



  j = start;

  {
    int ret = newStack (&now);

    if (ret != NORMAL)
      return ret;

    now.stack = a1;
    ret = newStack (&next);

    if (ret != NORMAL)
      return ret;
    next.stack = a2;
  }

  push (&now, fsm[0].next1);

  for (;;)
  {
    for (pos = 0; pos < now.size; pos++)
    {
      state = now.stack[pos];

      if (state == 0)
      {
	finished = j - start;
	continue;
      }

      if (fsm[state].command == NULL_CMD)
      {
	if (!is_present (&now, fsm[state].next2))
	  push (&now, fsm[state].next2);

	if (fsm[state].next1 != fsm[state].next2)
	{
	  if (!is_present (&now, fsm[state].next1))
	    push (&now, fsm[state].next1);
	}
      }

      else if (text[j] == '\0');
      /* faz com que \0 seja ignorado */

      else if (((fsm[state].command == CHAR_CMD) &&
		(fsm[state].ch == text[j])) ||
	       ((fsm[state].command == ANY_CMD) &&
                (text[j] != '\n')))
	push (&next, fsm[state].next1);

      else if (fsm[state].command == SPC_CMD)
	switch ((*(functions[(int) fsm[state].ch].function)) (text, j))
	{
	case no_match:
	  break;

	case normal_match:
	  push (&next, fsm[state].next1);
	  break;

	case no_char_match:
	  push (&now, fsm[state].next1);
	  break;
	}

      else if (fsm[state].command == CLASS_CMD)
      {
	int temp, found = 0, negate;

	temp = fsm[state].next1;
	negate = fsm[state].next2;
	state++;

	while (fsm[state].command != NULL_CMD)
	{
	  if (fsm[state].command == CLASS_CMD_RANGE)
	  {
	    if ((text[j] >= fsm[state].ch) &&
		(text[j] <= fsm[state].next1))
	    {
	      found = 1;
	      break;
	    }
	  }

	  else if ((fsm[state].command == CLASS_CMD_CHAR) &&
		   (text[j] == fsm[state].ch))
	  {
	    found = 1;
	    break;
	  };
	  state++;
	}

	if (found ^ negate)
	{
	  push (&next, temp);
	  state = temp;
	}
      }
      else if (fsm[state].command == BEGIN_CMD)
      {
	if (text[j - 1] == '\n' || j == 0)
	  push (&now, fsm[state].next1);
      }
      else if (fsm[state].command == END_CMD)
      {
	if (text[j] == '\n' || text[j] == '\0')
	  push (&now, fsm[state].next1);
      }

      /* codigo de teste -- imprime stack e queue */
      /* {
         int a;

         printf ("\nstack:");
         for (a = 0; a < now.size; a++)
         printf (" %d", now.stack[a]);

         printf ("\nqueue: ");
         for (a = 0; a < next.size; a++)
         printf (" %d", next.stack[a]);

         printf ("\n");
         } */
    }

    j++;

    if (next.size == 0)
    {
      clearStack (&now);
      clearStack (&next);

      if (finished > NOMATCH)
	return finished;

      return NOMATCH;
    }

    moveStack (&now, &next);
  }
}

/******************************************************/
/* Decide qual funcao percorrera a maquina de estados */
/* (recursiva ou nao-recursiva)                       */
/******************************************************/

long pm_match
  (char *text, fsm_st * fsm, long start, match_func function, void *user)
{
  long ret;
  short tested[10000];		/* vai ser eliminado */
  vars_st vars;

  /* se nao houver capturas no padrao, chama funcao nao-recursiva */

  if (fsm[0].ch == NOCAPTURE)
    return nmatch (text, fsm, start);

  /* dados usados pela funcao recursiva */

  vars.text = text;
  vars.fsm = fsm;
  vars.tested = tested;
  vars.function = function;

  vars.user = user;

  ret = rmatch (&vars, start, fsm[0].next1, NULL, 0);

  return (ret >= start) ? ret - start : ret;
}

/**************************************
 * retorna proxima capture disponivel *
 **************************************/

int pm_getCapture (int *which_one, long *start, long *size)
{
  if (capture_list.pos == capture_list.size)
    return 0;

  if(which_one != NULL)
    *which_one = capture_list.captures[capture_list.pos].which_one;
  if(start != NULL)
    *start = capture_list.captures[capture_list.pos].start;
  if(size != NULL)
    *size = capture_list.captures[capture_list.pos].size;

  capture_list.pos++;

  return 1;
}

/***********************
 * funcao de interface * 
 ***********************/

long pm_pmatch (long *first,
		char *text, char *pattern, long start,
		match_func function, void *user)
{
  fsm_st *fsm;
  long length, first1;
  enum bool is_recursive = false;
  static char last_pattern[MAX_LAST_PATTERN] = "";
  static fsm_st *last_fsm = NULL;

  /* se a chamada nao for recursiva, indica que a proxima chamada
     podera ser e que a chamada atual ainda nao e */

  first1 = start;

  if (will_be_recursive == false)
    will_be_recursive = true;
  else				/* esta dentro de uma chamada recursiva */
    is_recursive = true;

  /* se o texto nao for o mesmo ou se nao houver last_fsm */

  if (strcmp (pattern, last_pattern) || (last_fsm == NULL))
  {
    long result;

    /* pula caracteres de controle do pattern matching */

    if ((pattern[0] == '@') || (pattern[0] == '%'))
      result = pm_parse (pattern + 1, &fsm);

    else
      result = pm_parse (pattern, &fsm);

    if (result != PARSE_OK)
      return result;

    /* se texto novo couber no array last_pattern */

    if (strlen (pattern) < MAX_LAST_PATTERN)
    {
      /* se a chamada nao for recursiva, libera a maquina anterior */

      if ((is_recursive == false) && (last_fsm != NULL))
	free (last_fsm);

      /* guarda a nova maquina e o novo texto */

      last_fsm = fsm;
      strcpy (last_pattern, pattern);
    }

    /* verifica se usuario quer que a funcao percorra a string
       ou se so quer procurar a partir da posicao inicial */

    if (pattern[0] == '%')
    {
      long size = strlen (text);
      long first_length = NOMATCH;

      do
      {
	length = pm_match (text, fsm, start, function, user);

	if (length != NOMATCH)
	{
	  if (fsm[0].ch == NOCAPTURE)
	    pm_storeCaptures (0, start, start + length - 1, text, NULL);

	  if (first_length == NOMATCH)
	  {
	    first_length = length;
	    first1 = start;
	  }

	  start += length + 1;
	}
	else
	  start++;
      }
      while (start < size);

      length = first_length;	/* retorna o tamanho do primeiro match */
    }

    else if (pattern[0] != '@')
    {
      long size = strlen (text);

      do
      {
	length = pm_match (text, fsm, start, function, user);

	if (length != NOMATCH)
	{
	  if (fsm[0].ch == NOCAPTURE)
	    pm_storeCaptures (0, start, start + length - 1, text, NULL);

	  first1 = start;
	  break;
	}
	else
	  start++;
      }
      while (start < size);
    }

    else
      length = pm_match (text, fsm, start, function, user);

    /* se a chamada for recursiva, libera a maquina alocada
       para evitar "blocos" perdidos */

    if (is_recursive == true)
    {
      free (fsm);
      last_fsm = NULL;
    }
  }

  else
    /* utiliza a maquina anteriormente alocada */
  {
    fsm = last_fsm;

    if (pattern[0] == '%')
    {
      long size = strlen (text);
      long first_length = NOMATCH;

      do
      {
	length = pm_match (text, fsm, start, function, user);

	if (length != NOMATCH)
	{
	  if (fsm[0].ch == NOCAPTURE)
	    pm_storeCaptures (0, start, start + length - 1, text, NULL);

	  if (first_length == NOMATCH)
	  {
	    first_length = length;
	    first1 = start;
	  }

	  start += length + 1;
	}
	else
	  start++;
      }
      while (start < size);

      length = first_length;	/* retorna o tamanho do primeiro match */
    }

    else if (pattern[0] == '@')
    {
      long size = strlen (text);

      do
      {
	length = pm_match (text, fsm, start, function, user);

	if (length != NOMATCH)
	{
	  if (fsm[0].ch == NOCAPTURE)
	    pm_storeCaptures (0, start, start + length - 1, text, NULL);

	  first1 = start;
	  break;
	}
	else
	  start++;
      }
      while (start < size);
    }

    else
      length = pm_match (text, fsm, start, function, user);

  }
  /* se a chamada atual nao foi recursiva, a proxima nao o sera */

  if (is_recursive == false)
    will_be_recursive = false;

  if (first != NULL)
    *first = first1;

  return length;
}
